"""Fake in-memory database interface for demo purpose"""

import logging
from collections.abc import AsyncGenerator

from authlib.oauth2.rfc6749 import OAuth2Token
from jwt import PyJWTError

from sqlmodel import SQLModel, create_engine, select
from sqlmodel.ext.asyncio.session import AsyncSession
from sqlalchemy.ext.asyncio import create_async_engine

from oidc_test.auth import provider
from oidc_test.settings import settings
from oidc_test.auth.provider import Provider
from oidc_test.models import User, Role, Token
from oidc_test.auth_providers import providers

logger = logging.getLogger("oidc-test")

engine = create_async_engine(settings.db.sqla_url)
sync_engine = create_engine(settings.db.sqla_url)


class UserNotInDB(Exception):
    pass


class TokenNotInDb(Exception):
    pass


async def get_db_session() -> AsyncGenerator[AsyncSession]:
    async with AsyncSession(engine) as db_session:
        yield db_session


class Database:
    users: dict[str, User] = {}
    # TODO: key of the token table should be provider: sid
    tokens: dict[str, OAuth2Token] = {}

    # Last sessions for the user (key: users's subject id (sub))

    async def add_user(
        self,
        sub: str,
        auth_provider: Provider,
        token: OAuth2Token,
        # access_token: str,
        access_token_decoded: dict | None = None,
    ) -> User:
        if access_token_decoded is None:
            assert auth_provider.name is not None
            provider = providers[auth_provider.id]
            try:
                access_token_decoded = provider.decode(token["access_token"])
            except PyJWTError:
                access_token_decoded = {}
        user_info: dict = token["user_info"]
        sub = user_info["sub"]
        user = User(auth_provider_id=auth_provider.id, **user_info)
        user.userinfo = user_info
        # user.access_token = access_token
        # user.access_token_decoded = access_token_decoded
        # Add roles provided in the access token
        roles = set()
        try:
            r = access_token_decoded["resource_access"][auth_provider.client_id]["roles"]
            roles.update(r)
        except KeyError:
            pass
        try:
            r = access_token_decoded["realm_access"]["roles"]
            if isinstance(r, str):
                roles.add(r)
            else:
                roles.update(r)
        except KeyError:
            pass
        # user.roles = [Role(name=role_name) for role_name in roles]
        user.roles = []
        self.users[sub] = user
        return user

    async def get_user(self, sub: str, db_session: AsyncSession) -> User:
        query = select(User).where(User.sub == sub)
        user = (await db_session.exec(query)).first()
        if user is None:
            raise UserNotInDB
        return user

    async def get_or_add_user(
        self, sub: str, db_session: AsyncSession, auth_provider: Provider, token: OAuth2Token
    ):
        if user := self.get_user(sub, db_session):
            return user
        else:
            return await self.add_user(sub=sub, auth_provider=auth_provider, token=token)

    async def add_token(
        self, provider: Provider, token: OAuth2Token, db_session: AsyncSession
    ) -> None:
        """Store a token using as key the sid (auth provider's session id)
        in the id_token"""
        sid = provider.get_session_key(token["userinfo"])
        if existing_token := await db_session.get(Token, sid):
            # The token already exists: update it
            # XXX: check is token is different?
            existing_token.token = token
            db_session.add(existing_token)
            await db_session.commit()
        else:
            token = Token(sid=sid, token=token)
            db_session.add(token)
            await db_session.commit()

    async def get_token(
        self, provider: Provider, sid: str | None, db_session: AsyncSession
    ) -> OAuth2Token:
        # TODO: key of the token table should be provider: sid
        assert isinstance(provider, Provider)
        if sid is None:
            raise TokenNotInDb
        if token := await db_session.get(Token, sid):
            return OAuth2Token.from_dict(token.token)
        else:
            raise TokenNotInDb


async def create_db(drop=False):
    logger.debug(f"Connect to database with config: {settings.db}")
    async with engine.begin() as conn:
        if drop:
            await conn.run_sync(SQLModel.metadata.drop_all)
        await conn.run_sync(SQLModel.metadata.create_all)


db = Database()
